/* Do not change this notice!

   vg_renu.js : Web-based, interactive, Molecule Viewer in JavaScript

   (c) V.Ganesh. All rights reserved.

   Author: V.Ganesh
   Version: 1.0 (7th Nov 2004)

   LICENSE: LGPL
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License (LGPL) as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA,
or see http://www.gnu.org/copyleft/lesser.html 
 */

var mol = null;

function renuMouseMode(theOptions) {
  if (mol == null) return;

  mol.mouseMode = theOptions.options[theOptions.selectedIndex].value;
}

function renuShowLabel(showTheLabel) {
  if (mol == null) return;

  mol.showLabel = showTheLabel.checked;
  mol.draw();
}

function updateMolecule(canvas, theText, mouseMode, showLabel) {
   if (mol != null) mol.jg.clear();

   mol = new Molecule("New Molecule", canvas);
   mol.readFromTextArea(theText);
   mol.buildConnectivity();
   mol.draw();

   renuMouseMode(mouseMode);
   renuShowLabel(showLabel);
}

function renuMouseUp(event) {
  if (mol == null) return;

  mol.onMouseUp(event);
}

function renuMouseDown(event) {
  if (mol == null) return;

  mol.onMouseDown(event);
}

function renuMouseOver(event) {
  if (mol == null) return;

  mol.onMouseOver(event);
}

function renuMouseMove(event) {
  if (mol == null) return;

  mol.onMouseMove(event);
}

function renuMouseOut(event) {
  if (mol == null) return;

  mol.onMouseOut(event);
}

function Matrix3D() {
        this.xx = this.xy = this.xz = this.xo = 0.0;
        this.yx = this.yy = this.yz = this.yo = 0.0;
        this.zx = this.zy = this.zz = this.zo = 0.0;

        // make it unit               
        this.xx = 1.0;
        this.yy = 1.0;
        this.zz = 1.0;

        // Scale by f in all dimensions 
        this.scale = function(f) {
            this.xx *= f;
            this.xy *= f;
            this.xz *= f;
            this.xo *= f;
            this.yx *= f;
            this.yy *= f;
            this.yz *= f;
            this.yo *= f;
            this.zx *= f;
            this.zy *= f;
            this.zz *= f;
            this.zo *= f;
        }
        
        // Scale along each axis independently 
        this.scale = function(xf, yf, zf) {
            this.xx *= xf;
            this.xy *= xf;
            this.xz *= xf;
            this.xo *= xf;
            this.yx *= yf;
            this.yy *= yf;
            this.yz *= yf;
            this.yo *= yf;
            this.zx *= zf;
            this.zy *= zf;
            this.zz *= zf;
            this.zo *= zf;
        }
        
        // Translate the origin 
        this.translate = function(x, y, z) {
            this.xo += x;
            this.yo += y;
            this.zo += z;
        }
        
        // rotate theta degrees about the y axis 
        this.yrot = function(theta) {
            theta *= (Math.PI / 180);
            var ct = Math.cos(theta);
            var st = Math.sin(theta);

            var Nxx = (this.xx * ct + this.zx * st);
            var Nxy = (this.xy * ct + this.zy * st);
            var Nxz = (this.xz * ct + this.zz * st);
            var Nxo = (this.xo * ct + this.zo * st);

            var Nzx = (this.zx * ct - this.xx * st);
            var Nzy = (this.zy * ct - this.xy * st);
            var Nzz = (this.zz * ct - this.xz * st);
            var Nzo = (this.zo * ct - this.xo * st);

            this.xo = Nxo;
            this.xx = Nxx;
            this.xy = Nxy;
            this.xz = Nxz;
            this.zo = Nzo;
            this.zx = Nzx;
            this.zy = Nzy;
            this.zz = Nzz;
        }
        
        // rotate theta degrees about the x axis 
        this.xrot = function(theta) {
	    theta *= (Math.PI / 180);	
            var ct = Math.cos(theta);
            var st = Math.sin(theta);

            var Nyx = (this.yx * ct + this.zx * st);
            var Nyy = (this.yy * ct + this.zy * st);
            var Nyz = (this.yz * ct + this.zz * st);
            var Nyo = (this.yo * ct + this.zo * st);

            var Nzx = (this.zx * ct - this.yx * st);
            var Nzy = (this.zy * ct - this.yy * st);
            var Nzz = (this.zz * ct - this.yz * st);
            var Nzo = (this.zo * ct - this.yo * st);

            this.yo = Nyo;
            this.yx = Nyx;
            this.yy = Nyy;
            this.yz = Nyz;
            this.zo = Nzo;
            this.zx = Nzx;
            this.zy = Nzy;
            this.zz = Nzz;
        }
        
        // rotate theta degrees about the z axis 
        this.zrot = function(theta) {
	    theta *= (Math.PI / 180);	
            var ct = Math.cos(theta);
            var st = Math.sin(theta);

            var Nyx = (this.yx * ct + this.xx * st);
            var Nyy = (this.yy * ct + this.xy * st);
            var Nyz = (this.yz * ct + this.xz * st);
            var Nyo = (this.yo * ct + this.xo * st);

            var Nxx = (this.xx * ct - this.yx * st);
            var Nxy = (this.xy * ct - this.yy * st);
            var Nxz = (this.xz * ct - this.yz * st);
            var Nxo = (this.xo * ct - this.yo * st);

            this.yo = Nyo;
            this.yx = Nyx;
            this.yy = Nyy;
            this.yz = Nyz;
            this.xo = Nxo;
            this.xx = Nxx;
            this.xy = Nxy;
            this.xz = Nxz;
        }
        
        // Multiply this matrix by a second: M = M*R 
        this.mult = function(rhs) {
            var lxx = this.xx * rhs.xx + this.yx * rhs.xy + this.zx * rhs.xz;
            var lxy = this.xy * rhs.xx + this.yy * rhs.xy + this.zy * rhs.xz;
            var lxz = this.xz * rhs.xx + this.yz * rhs.xy + this.zz * rhs.xz;
            var lxo = this.xo * rhs.xx + this.yo * rhs.xy + this.zo * rhs.xz + rhs.xo;

            var lyx = this.xx * rhs.yx + this.yx * rhs.yy + this.zx * rhs.yz;
            var lyy = this.xy * rhs.yx + this.yy * rhs.yy + this.zy * rhs.yz;
            var lyz = this.xz * rhs.yx + this.yz * rhs.yy + this.zz * rhs.yz;
            var lyo = this.xo * rhs.yx + this.yo * rhs.yy + this.zo * rhs.yz + rhs.yo;

            var lzx = this.xx * rhs.zx + this.yx * rhs.zy + this.zx * rhs.zz;
            var lzy = this.xy * rhs.zx + this.yy * rhs.zy + this.zy * rhs.zz;
            var lzz = this.xz * rhs.zx + this.yz * rhs.zy + this.zz * rhs.zz;
            var lzo = this.xo * rhs.zx + this.yo * rhs.zy + this.zo * rhs.zz + rhs.zo;

            this.xx = lxx;
            this.xy = lxy;
            this.xz = lxz;
            this.xo = lxo;
            
            this.yx = lyx;
            this.yy = lyy;
            this.yz = lyz;
            this.yo = lyo;

            this.zx = lzx;
            this.zy = lzy;
            this.zz = lzz;
            this.zo = lzo;
        }

        // Reinitialize to the unit matrix 
        this.unit = function() {
            this.xo = 0;
            this.xx = 1;
            this.xy = 0;
            this.xz = 0;
            this.yo = 0;
            this.yx = 0;
            this.yy = 1;
            this.yz = 0;
            this.zo = 0;
            this.zx = 0;
            this.zy = 0;
            this.zz = 1;
        }        

        // transform the atoms into a new space
        this.transform = function(atoms) {
            var tAtoms = new Array(atoms.length);
            
            for(var i=0; i<atoms.length; i++) {
               var x = atoms[i].x;
               var y = atoms[i].y;
               var z = atoms[i].z;

	       tAtoms[i] = new Atom(atoms[i].symbol, 
                Math.round(x * this.xx + y * this.xy + z * this.xz + this.xo),
                Math.round(x * this.yx + y * this.yy + z * this.yz + this.yo),
                Math.round(x * this.zx + y * this.zy + z * this.zz + this.zo));
            } // end for
 
            return tAtoms;
	}
} // end of object Matrix3D

function Atom(symbol, x, y, z) {
        this.symbol = symbol;
        this.x = x;
        this.y = y;
        this.z = z;
        this.connectedList = new Array();
} // end of object Atom

function Molecule(title, canvas) {  
        this.title = title;

        // graphics object
        this.jg    = new jsGraphics(canvas);
        this.jg.setPrintable(true);

        // width / height
        this.canvasWidth = this.jg.cnv.clientWidth;
        this.canvasHeight = this.jg.cnv.clientHeight;

        // mouse props
        this.prevx = 0;
        this.prevy = 0;
        this.mousePressed = false;

        // mouse mode
        this.mouseMode = "rotate";

        // the even handlers

        // mouse down
        this.onMouseDown = function(event) {
           this.prevx = event.x;
           this.prevy = event.y;       
           this.mousePressed = true; 
        }

        // mouse up
        this.onMouseUp = function(event) {
            var x = event.x;
            var y = event.y;
            
            this.tmat.unit();
            
            if (this.mouseMode == "rotate") {
              var xtheta = (this.prevy - y) * (360.0 / this.canvasWidth);
              var ytheta = (x - this.prevx) * (360.0 / this.canvasHeight);
            
              this.tmat.xrot(xtheta);
              this.tmat.yrot(ytheta);
              this.amat.mult(this.tmat);            
            } else if (this.mouseMode == "zoom") {
              if ((this.prevx < x) || (this.prevy < y)) {
                 this.scalefudge += 0.1;
              } else {
                 this.scalefudge -= 0.1;
              } // end if
            } // end if

            this.prevx = x;
            this.prevy = y;      

            this.mousePressed = false;     

            this.draw();
        }

        // mouse move
        this.onMouseMove = function(event) {
            if (this.mousePressed) {
               var x = event.x;
               var y = event.y;
            
               this.tmat.unit();
            
               if (this.mouseMode == "rotate") {
                 var xtheta = (this.prevy - y) * (360.0 / this.canvasWidth);
                 var ytheta = (x - this.prevx) * (360.0 / this.canvasHeight);
            
                 this.tmat.xrot(xtheta);
                 this.tmat.yrot(ytheta);
                 this.amat.mult(this.tmat);
               } else if (this.mouseMode == "zoom") {
                 if ((this.prevx < x) || (this.prevy < y)) {
                    this.scalefudge += 0.1;
                 } else {
                    this.scalefudge -= 0.1;
                 } // end if
               } // end if
            
               this.prevx = x;
               this.prevy = y;

               this.draw();
            } // end if
        }

        // mouse over
        this.onMouseOver = function(event) {
        }

        // mouse out
        this.onMouseOut = function(event) {
        }

        // ...
        this.atoms = new Array();
        this.mat   = new Matrix3D(); // final matrix
        this.amat  = new Matrix3D(); // intermediate matrix
        this.tmat  = new Matrix3D(); // translation...
        this.covalentRadius = new Array();
        this.covalentRadius["H"]  = 0.23;	
        this.covalentRadius["C"]  = 0.77;
        this.covalentRadius["N"]  = 0.75;
        this.covalentRadius["O"]  = 0.73;
        this.covalentRadius["S"]  = 1.02;
        this.covalentRadius["P"]  = 1.06;
        this.covalentRadius["NA"] = 1.54;  // Na
        this.covalentRadius["CL"] = 0.99;  // Cl
        this.covalentRadius["CA"] = 1.74;  // Ca
      
        // the color table:
        this.colorTable = new Array();            
        this.colorTable["H"]  = "gray";
        this.colorTable["C"]  = "darkGray";
        this.colorTable["N"]  = "blue";
        this.colorTable["O"]  = "red";
        this.colorTable["S"]  = "yellow";
        this.colorTable["P"]  = "orange";
        this.colorTable["NA"] = "pink";   // Na
        this.colorTable["CL"] = "green";  // Cl
        this.colorTable["CA"] = "white";  // Ca
  
        // the bounding box
        this.xmax = 0.0;
        this.xmin = 0.0;
        this.ymax = 0.0;
        this.ymin = 0.0;
        this.zmax = 0.0;
        this.zmin = 0.0;

        // the scale factor
        this.scalefudge = 0.8;

        // show label ? , default yes
        this.showLabel = true;

        // add an atom to molecule
        this.addAtom = function(atom) {
            this.atoms.push(atom);
        }

        // read from text area
        this.readFromTextArea = function(textArea) {
            var lines = textArea.innerText.split(/\n/);

            var numberOfAtoms = parseInt(lines[0]);
            this.title        = lines[1];
             
            for(var i = 2; i<numberOfAtoms+2; i++) {
               var words = lines[i].split(/\s/);

               this.addAtom(new Atom(words[0], 
                                     parseFloat(words[1]), 
                                     parseFloat(words[2]), 
                                     parseFloat(words[3])));
            } // end for
        }

        // build the connectivity
        this.buildConnectivity = function() {
             for(var i=0; i<this.atoms.length; i++) {
                for(var j=0; j<i; j++) {
                   if (this.strongBondExist(this.atoms[i], this.atoms[j])) {
                      this.atoms[i].connectedList.push(j);
                      this.atoms[j].connectedList.push(i);
                   } // end if
                } // end for
             } // end for

             // find bounding box
             this.findBB();
	}

        // find the bounding box
        this.findBB = function() {
            var xmin = this.atoms[0].x;
            var xmax = xmin;
            var ymin = this.atoms[0].y;
            var ymax = ymin;
            var zmin = this.atoms[0].z;
            var zmax = zmin;
            var x, y, z;

            for (var i=1; i<this.atoms.length; i++) {
                x = this.atoms[i].x;
                
                if (x < xmin)
                    xmin = x;
                if (x > xmax)
                    xmax = x;
                
                y = this.atoms[i].y;
                
                if (y < ymin)
                    ymin = y;
                if (y > ymax)
                    ymax = y;
                
                z = this.atoms[i].z;
                
                if (z < zmin)
                    zmin = z;
                if (z > zmax)
                    zmax = z;
            } // end for
            
            this.xmax = xmax;
            this.xmin = xmin;
            this.ymax = ymax;
            this.ymin = ymin;
            this.zmax = zmax;
            this.zmin = zmin;
        }

        // does a strong bond exist between two atoms?
        this.strongBondExist = function(a1, a2) {
            var radiiSum = this.covalentRadius[a1.symbol.toUpperCase()]
                              + this.covalentRadius[a2.symbol.toUpperCase()];
                              
            var distance = Math.sqrt(((a2.x-a1.x) * (a2.x-a1.x))
                                    + ((a2.y-a1.y) * (a2.y-a1.y))
                                    + ((a2.z-a1.z) * (a2.z-a1.z)));

            // taken from cambridge cluster database
            return (((radiiSum - 0.4) < distance) 
                   && (distance < (radiiSum + 0.4)));
	}

        // draw this molecule
        this.draw = function() {
            var xw = this.xmax - this.xmin;
            var yw = this.ymax - this.ymin;
            var zw = this.zmax - this.zmin;
                
            if (yw > xw)
                xw = yw;
            if (zw > xw)
                xw = zw;
                
            var f1 = this.canvasWidth / xw;
            var f2 = this.canvasHeight / xw;
                
            var xfac = 0.7;
             
            if (f1 < f2) xfac *= f1;
            else         xfac *= f2; 
            
            xfac *= this.scalefudge; 

            this.mat.unit();
            this.mat.translate(-(this.xmin + this.xmax) / 2.0,
                               -(this.ymin + this.ymax) / 2.0,
                               -(this.zmin + this.zmax) / 2.0);
            this.mat.mult(this.amat);
            
            this.mat.scale(xfac, -xfac, 16 * xfac / this.canvasWidth);
            this.mat.translate(this.canvasWidth / 2, this.canvasHeight / 2, 8);
                
            var tAtoms = this.mat.transform(this.atoms);
            var atom1, atom2, mid;

            this.jg.clear();

            for(var i=0; i<this.atoms.length; i++) {
                atom1 = tAtoms[i];

                if (this.showLabel && !this.mousePressed) {
                    this.jg.setColor(this.colorTable[atom1.symbol]);
                    this.jg.drawString(atom1.symbol, atom1.x, atom1.y);
                } // end if

                for(var j=0; j<this.atoms[i].connectedList.length; j++) { 
                    atom2 = tAtoms[this.atoms[i].connectedList[j]]; 

                    mid = new Atom("", Math.round((atom1.x + atom2.x) / 2),
                                       Math.round((atom1.y + atom2.y) / 2),
                                       0); // neglet z!

                    this.jg.setColor(this.colorTable[atom1.symbol]);
                    this.jg.drawLine(atom1.x, atom1.y, mid.x, mid.y);
                    this.jg.setColor(this.colorTable[atom2.symbol]);
                    this.jg.drawLine(atom2.x, atom2.y, mid.x, mid.y);
                } // end for
            } // end for

            this.jg.paint();
        }
} // end of object Molecule
